==> jsonstreamreader-1.1.6.tar.gz: lib/source/remote/remotesource.dart <==
import 'package:path/path.dart';
import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:math';

import 'remotescheme.dart';

import 'resourcestate.dart';
import 'filestateprovider.dart';
import 'remotemethod.dart';

///Used for Remote Sources remote file or api
class RemoteSource {
  File _file;
  String _url;
  Uri _uri;
  int _totalSize = 0;
  int _currentSize = 0;
  bool _completed = false;
  RemoteScheme _scheme = RemoteScheme.HTTP;
  RemoteMethod _method = RemoteMethod.GET;
  Map<String, String> _body = new Map<String, String>();
  Map<String, String> _headers = new Map<String, String>();
  Map<String, String> _secondaryheaders = new Map<String, String>();
  FileStateProvider sp;
  StreamController<Null> _downloading = new StreamController<Null>();
  bool _loading = false;
  HttpClientResponse response;
  bool _supportsRange = false;
  bool _compressedResponse = true;
  bool _paused = false;
  bool _trim = false;
  int _offset = -1;
  StreamController<Null> _download = new StreamController<Null>.broadcast();

  ///Create RemoteSource instance from a string
  RemoteSource.fromString(String content) {
    Map<String, dynamic> data = jsonDecode(content);
    this._url = data['uri'];
    this._supportsRange = data['supportsRange'];
    this._trim = data['trim'];
    this._compressedResponse = data['compressedResponse'];
    this._offset = data['offset'];
    Uri uriBefore = Uri.parse(this._url);
    this._body = uriBefore.queryParameters;
    if (uriBefore.scheme.toLowerCase() == 'http') {
      this._uri = Uri.http(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    } else {
      this._scheme = RemoteScheme.HTTPS;
      this._uri = Uri.https(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    }
    this.setMethod(RemoteMethod.GET.toString() == data['method'].toString()
        ? RemoteMethod.GET
        : RemoteMethod.POST);
    data['headers'].forEach((String key, dynamic val) {
      this.addHeader(key, val.toString());
    });
    Map<String, String> newBody = new Map<String, String>();
    data['body'].forEach((String key, dynamic val) {
      newBody[key] = val.toString();
    });
    this.setBody(newBody);
    sp = new FileStateProvider();
  }
  RemoteSource(String url,
      {bool supportsRange = false,
      bool compressedResponse = true,
      bool trim = false,
      int offset = 0}) {
    this._url = url;
    this._supportsRange = supportsRange;
    this._compressedResponse = compressedResponse;
    this._trim = trim;
    this._offset = offset;
    Uri uriBefore = Uri.parse(this._url);
    if (uriBefore.scheme.toLowerCase() == 'http') {
      this._uri = Uri.http(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    } else {
      this._scheme = RemoteScheme.HTTPS;
      this._uri = Uri.https(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    }
    sp = new FileStateProvider();
  }
  String get path {
    return jsonEncode({
      'method': this._method.toString(),
      'uri': this._uri.origin + this._uri.path,
      'headers': this._headers,
      'body': this._body,
      'supportsRange': this._supportsRange,
      'compressedResponse': this._compressedResponse,
      'offset': this._offset,
      'trim': this._trim,
      'last': this._currentSize
    });
  }

  //get the current size of the source
  int getCurrent() {
    return this._currentSize;
  }

  //if completed get the total size of the file
  int getTotalSize() {
    return this._totalSize;
  }

  //has the file finished downloading
  bool isCompleted() {
    return this._completed;
  }

  //delete this resource currently stored in temp folder
  Future<FileSystemEntity> delete() async {
    return this._file.delete();
  }

  //set the body of the request
  void setBody(Map<String, String> body) {
    this._body = body;
    if (this._scheme == RemoteScheme.HTTP) {
      this._uri = Uri.http(this._uri.authority, this._uri.path, this._body);
    } else {
      this._uri = Uri.https(this._uri.authority, this._uri.path, this._body);
    }
  }

  //add a request header
  void addHeader(String property, String value) {
    if (property.toLowerCase().trim() ==
            HttpHeaders.acceptEncodingHeader.toLowerCase().trim() ||
        property.toLowerCase().trim() ==
            HttpHeaders.contentEncodingHeader.toLowerCase().trim()) {
      throw 'Cannot Set Accept-Encoding or Content-Encoding';
    }
    this._headers[property] = value;
  }

  //add a secondary request header that is deleted when download is called
  //for the first time.NB this will not be used again if download is to be resumed
  void addSecondaryHeader(String property, String value) {
    if (property.toLowerCase().trim() ==
            HttpHeaders.acceptEncodingHeader.toLowerCase().trim() ||
        property.toLowerCase().trim() ==
            HttpHeaders.contentEncodingHeader.toLowerCase().trim()) {
      throw 'Cannot Set Accept-Encoding or Content-Encoding';
    }
    this._secondaryheaders[property] = value;
  }

  //set the method of request either post or get
  void setMethod(RemoteMethod method) {
    this._method = method;
  }

  //set the scheme of request either http or https
  void setScheme(RemoteScheme protocol) {
    this._scheme = protocol;
    if (this._scheme == RemoteScheme.HTTP) {
      this._uri = Uri.http(
          this._uri.authority, this._uri.path, this._uri.queryParameters);
    } else {
      this._uri = Uri.https(
          this._uri.authority, this._uri.path, this._uri.queryParameters);
    }
  }

  //get the length of the file from request header
  Future<int> length() {
    Completer<int> c = new Completer<int>();
    if (this._totalSize > 0) {
      new Timer(new Duration(microseconds: 0), () {
        c.complete(this._totalSize);
      });
    } else {
      new HttpClient().headUrl(this._uri).then((HttpClientRequest request) {
        request.headers
            .set(HttpHeaders.acceptEncodingHeader, 'accept-encoding=identity');
        this._headers.forEach((String property, String value) {
          request.headers.set(property, value);
        });
        this._secondaryheaders.forEach((String property, String value) {
          request.headers.set(property, value);
        });
        return request.close();
      }).then((HttpClientResponse response) {
        try {
          this._totalSize = int.parse(response.headers['content-length'][0]);
          c.complete(this._totalSize);
        } catch (e) {
          this._totalSize = 0;
          c.complete(this._totalSize);
        }
      }).catchError((error) {});
    }
    return c.future;
  }

  //get the file temporary stored
  File getFile() {
    return this._file;
  }

  //Stream indicator for when download is completed
  Stream<Null> get downloaded {
    return _download.stream;
  }

  //pause download,without range header download is restart when resume is called
  Future<void> pause() async {
    Completer<void> completer = new Completer<void>();
    if (this._loading) {
      new Timer(new Duration(microseconds: 0), () {
        this._paused = true;
        this.response.detachSocket();
      });
    } else {
      _downloading.stream.listen((d) {
        this._paused = true;
        this.response.detachSocket();
      });
    }
    return completer.future;
  }

  //resume download
  void resume() {
    this._paused = false;
    this.download();
  }

//download resource to temp folder
  void download() async {
    Uri uri = Uri.parse(this._url);
    String pa = uri.path;
    if (pa.isEmpty) {
      var random = new Random();
      pa = "file_" + random.nextInt(10000).toString();
    }
    String base = basename(uri.path);
    String filename = Directory.systemTemp.path +
        '/' +
        (base.contains('.') ? base : base + '.json');
    File downloadFile = new File(filename);
    if (this._supportsRange) {
      downloadFile.exists().then((bool exists) {
        if (exists) {
          this._file = downloadFile;
          downloadFile.length().then((int l) {
            this._currentSize = l;
            if (l == 0) {
              this._rangeDownload(downloadFile, 0);
            } else {
              this._rangeDownload(downloadFile, l + this._offset);
            }
          }).catchError((error) {
            _download.addError(error);
            _download.close();
          });
        } else {
          downloadFile.create().then((File f) {
            this._file = f;
            this._rangeDownload(f, 0);
          }).catchError((error) {
            _download.addError(error);
            _download.close();
          });
        }
      }).catchError((error) {
        _download.addError(error);
        _download.close();
      });
    } else {
      downloadFile.create().then((File f) {
        this._file = f;
        this._rangeDownload(f, 0);
      }).catchError((error) {
        _download.addError(error);
        _download.close();
      });
    }
  }

  void _rangeDownload(File f, int start) {
    FileMode fm = FileMode.write;
    if (start > 0) {
      fm = FileMode.append;
      this._secondaryheaders[HttpHeaders.rangeHeader] =
          "bytes=" + start.toString() + "-";
    }
    int i = 0;
    Future<HttpClientRequest> req;
    if (this._method == RemoteMethod.GET) {
      req = new HttpClient().getUrl(this._uri);
    } else {
      req = new HttpClient().postUrl(this._uri);
    }
    req.then((HttpClientRequest request) {
      this._headers.forEach((String property, String value) {
        request.headers.set(property, value);
      });
      this._secondaryheaders.forEach((String property, String value) {
        request.headers.set(property, value);
      });
      if (!this._compressedResponse) {
        request.headers
            .set(HttpHeaders.acceptEncodingHeader, 'accept-encoding=identity');
      }
      return request.close();
    }).then((HttpClientResponse response) {
      if (!this._loading && !_downloading.isClosed) {
        _downloading.add(null);
      }
      this._loading = true;
      this.response = response;
      response.listen((contents) {
        List<int> c = [];
        if (i > 0) {
          fm = FileMode.append;
        }
        if (i == 0 && this._trim) {
          c = contents.getRange(1, contents.length).toList();
          f.writeAsBytesSync(c, mode: fm);
          this._currentSize += c.length;
        } else {
          f.writeAsBytesSync(contents, mode: fm);
          this._currentSize += contents.length;
        }
        this.sp.pushState(this._currentSize, FileStateProvider.WAITING);
        i++;
      }, onDone: () {
        if (!_paused && !_download.isClosed) {
          this._secondaryheaders.remove(HttpHeaders.contentRangeHeader);
          this._secondaryheaders.remove(HttpHeaders.acceptRangesHeader);
          _downloading.close();
          this._loading = false;
          this._completed = true;
          this._totalSize = this._currentSize;
          this.sp.pushState(this._currentSize, FileStateProvider.COMPLETED);
          sp.close();
          _download.add(null);
          _download.close();
        }
        //sink.close();
        i = null;
      }, onError: (error) {
        _download.addError(error);
        _download.close();
      });
    });
  }

  //openRead implementation of a remote source
  Stream<List<int>> openRead(int start, [int end = 0]) {
    StreamController<List<int>> cc = new StreamController<List<int>>();
    if (end == 0 && this._completed) {
      this._file.openRead(start).listen((List<int> line) {
        cc.add(line);
      }, onDone: () {
        cc.close();
      }, onError: (e) {
        cc.addError(e);
      });
    } else if (end > 0 && end <= this._currentSize) {
      this._file.openRead(start, end).listen((List<int> line) {
        cc.add(line);
      }, onDone: () {
        cc.close();
      }, onError: (e) {
        cc.addError(e);
      });
    } else {
      if (end == 0) {
        this.sp.pushState(this._totalSize, FileStateProvider.CALLED);
      } else {
        this.sp.pushState(end, FileStateProvider.CALLED);
      }
      StreamSubscription<ResourceState> fstream;
      fstream = sp.ready().listen((ResourceState data) {
        if (end == 0) {
          if (this._completed) {
            this._file.openRead(start).listen((List<int> line) {
              if (!cc.isClosed) {
                cc.add(line);
              } else {
                fstream.cancel();
              }
            }, onDone: () {
              cc.close();
            }, onError: (e) {
              cc.addError(e);
            });
          }
        } else if (data.length >= end) {
          this._file.openRead(start, end).listen((List<int> line) {
            if (!cc.isClosed) {
              cc.add(line);
            } else {
              fstream.cancel();
            }
          }, onDone: () {
            cc.close();
          }, onError: (e) {
            cc.addError(e);
          });
        }
      });
    }
    return cc.stream;
  }
}

==> jsonstreamreader-1.1.8.tar.gz: lib/source/remote/remotesource.dart <==
import 'package:path/path.dart';
import 'dart:async';
import 'dart:convert';
import 'dart:io';
import 'dart:math';

import 'futureread.dart';
import 'remotescheme.dart';
import 'filestateprovider.dart';
import 'remotemethod.dart';

///Used for Remote Sources or remote files
class RemoteSource {
  File _file;
  String _filename;
  String _url;
  Uri _uri;
  int _totalSize = 0;
  int _currentSize = 0;
  bool _completed = false;
  RemoteScheme _scheme = RemoteScheme.HTTP;
  RemoteMethod _method = RemoteMethod.GET;
  Map<String, String> _body = new Map<String, String>();
  Map<String, String> _headers = new Map<String, String>();
  Map<String, String> _secondaryheaders = new Map<String, String>();
  FileStateProvider sp;
  StreamController<Null> _downloading = new StreamController<Null>();
  bool _loading = false;
  HttpClientResponse response;
  bool _supportsRange = false;
  bool _compressedResponse = true;
  bool _paused = false;
  bool _trim = false;
  int _offset = -1;
  StreamController<Null> _download = new StreamController<Null>.broadcast();

  ///Create RemoteSource instance from a string
  RemoteSource.fromString(String content) {
    Map<String, dynamic> data = jsonDecode(content);
    this._url = data['uri'];
    this._supportsRange = data['supportsRange'];
    this._trim = data['trim'];
    this._compressedResponse = data['compressedResponse'];
    this._offset = data['offset'];
    Uri uriBefore = Uri.parse(this._url);
    this._body = uriBefore.queryParameters;
    String pa = uriBefore.path;
    if (pa.isEmpty) {
      var random = new Random();
      pa = "file_" + random.nextInt(10000).toString();
    }
    String base = basename(uriBefore.path);
    this._filename = Directory.systemTemp.path +
        '/' +
        (base.contains('.') ? base : base + '.json');
    this._file = new File(this._filename);
    if (uriBefore.scheme.toLowerCase() == 'http') {
      this._uri = new Uri.http(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    } else {
      this._scheme = RemoteScheme.HTTPS;
      this._uri = new Uri.https(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    }
    this.setMethod(RemoteMethod.GET.toString() == data['method'].toString()
        ? RemoteMethod.GET
        : RemoteMethod.POST);
    data['headers'].forEach((String key, dynamic val) {
      this.addHeader(key, val.toString());
    });
    Map<String, String> newBody = new Map<String, String>();
    data['body'].forEach((String key, dynamic val) {
      newBody[key] = val.toString();
    });
    this.setBody(newBody);
    sp = new FileStateProvider();
  }
  RemoteSource(String url,
      {bool supportsRange = false,
      bool compressedResponse = true,
      bool trim = false,
      int offset = 0}) {
    this._url = url;
    this._supportsRange = supportsRange;
    this._compressedResponse = compressedResponse;
    this._trim = trim;
    this._offset = offset;
    Uri uriBefore = Uri.parse(this._url);
    String pa = uriBefore.path;
    if (pa.isEmpty) {
      var random = new Random();
      pa = "file_" + random.nextInt(10000).toString();
    }
    String base = basename(uriBefore.path);
    this._filename = Directory.systemTemp.path +
        '/' +
        (base.contains('.') ? base : base + '.json');
    this._file = new File(this._filename);
    if (uriBefore.scheme.toLowerCase() == 'http') {
      this._uri = new Uri.http(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    } else {
      this._scheme = RemoteScheme.HTTPS;
      this._uri = new Uri.https(
          uriBefore.authority, uriBefore.path, uriBefore.queryParameters);
    }
    sp = new FileStateProvider();
  }
  String get path {
    return jsonEncode({
      'method': this._method.toString(),
      'uri': this._uri.origin + this._uri.path,
      'headers': this._headers,
      'body': this._body,
      'supportsRange': this._supportsRange,
      'compressedResponse': this._compressedResponse,
      'offset': this._offset,
      'trim': this._trim,
      'last': this._currentSize
    });
  }

  //get the current size of the source
  int getCurrent() {
    return this._currentSize;
  }

  //if completed get the total size of the file
  int getTotalSize() {
    return this._totalSize;
  }

  //has the file finished downloading
  bool isCompleted() {
    return this._completed;
  }

  //delete this resource currently stored in temp folder
  Future<FileSystemEntity> delete() async {
    return this._file.delete();
  }

  //set the body of the request
  void setBody(Map<String, String> body) {
    this._body = body;
    if (this._scheme == RemoteScheme.HTTP) {
      this._uri = new Uri.http(this._uri.authority, this._uri.path, this._body);
    } else {
      this._uri =
          new Uri.https(this._uri.authority, this._uri.path, this._body);
    }
  }

  //add a request header
  void addHeader(String property, String value) {
    if (property.toLowerCase().trim() ==
            HttpHeaders.acceptEncodingHeader.toLowerCase().trim() ||
        property.toLowerCase().trim() ==
            HttpHeaders.contentEncodingHeader.toLowerCase().trim()) {
      throw 'Cannot Set Accept-Encoding or Content-Encoding';
    }
    this._headers[property] = value;
  }

  //add a secondary request header that is deleted when download is called
  //for the first time.NB this will not be used again if download is to be resumed
  void addSecondaryHeader(String property, String value) {
    if (property.toLowerCase().trim() ==
            HttpHeaders.acceptEncodingHeader.toLowerCase().trim() ||
        property.toLowerCase().trim() ==
            HttpHeaders.contentEncodingHeader.toLowerCase().trim()) {
      throw 'Cannot Set Accept-Encoding or Content-Encoding';
    }
    this._secondaryheaders[property] = value;
  }

  //set the method of request either post or get
  void setMethod(RemoteMethod method) {
    this._method = method;
  }

  //set the scheme of request either http or https
  void setScheme(RemoteScheme protocol) {
    this._scheme = protocol;
    if (this._scheme == RemoteScheme.HTTP) {
      this._uri = new Uri.http(
          this._uri.authority, this._uri.path, this._uri.queryParameters);
    } else {
      this._uri = new Uri.https(
          this._uri.authority, this._uri.path, this._uri.queryParameters);
    }
  }

  //get the length of the file from request header
  Future<int> length() {
    Completer<int> c = new Completer<int>();
    if (this._totalSize > 0) {
      new Timer(new Duration(microseconds: 0), () {
        c.complete(this._totalSize);
      });
    } else {
      new HttpClient().headUrl(this._uri).then((HttpClientRequest request) {
        request.headers
            .set(HttpHeaders.acceptEncodingHeader, 'accept-encoding=identity');
        this._headers.forEach((String property, String value) {
          request.headers.set(property, value);
        });
        this._secondaryheaders.forEach((String property, String value) {
          request.headers.set(property, value);
        });
        return request.close();
      }).then((HttpClientResponse response) {
        try {
          this._totalSize = int.parse(response.headers['content-length'][0]);
          c.complete(this._totalSize);
        } catch (e) {
          this._totalSize = 0;
          c.complete(this._totalSize);
        }
      }).catchError((error) {});
    }
    return c.future;
  }

  //get the file temporary stored
  File getFile() {
    return this._file;
  }

  //Stream indicator for when download is completed
  Stream<Null> get downloaded {
    return _download.stream;
  }

  //pause download,without range header download is restart when resume is called
  Future pause() async {
    Completer completer = new Completer();
    if (this._loading) {
      new Timer(new Duration(microseconds: 0), () {
        this._paused = true;
        this.response.detachSocket();
      });
    } else {
      _downloading.stream.listen((d) {
        this._paused = true;
        this.response.detachSocket();
      });
    }
    return completer.future;
  }

  //resume download
  void resume() {
    this._paused = false;
    this.download();
  }

//download resource to temp folder
  Future download() async {
    File downloadFile = this._file;
    if (this._supportsRange) {
      downloadFile.exists().then((bool exists) {
        if (exists) {
          this._file = downloadFile;
          downloadFile.length().then((int l) {
            this._currentSize = l;
            if (l == 0) {
              this._rangeDownload(downloadFile, 0);
            } else {
              this._rangeDownload(downloadFile, l + this._offset);
            }
          }).catchError((error) {
            _download.addError(error);
            _download.close();
          });
        } else {
          downloadFile.create().then((File f) {
            this._file = f;
            this._rangeDownload(f, 0);
          }).catchError((error) {
            _download.addError(error);
            _download.close();
          });
        }
      }).catchError((error) {
        _download.addError(error);
        _download.close();
      });
    } else {
      downloadFile.create().then((File f) {
        this._file = f;
        this._rangeDownload(f, 0);
      }).catchError((error) {
        _download.addError(error);
        _download.close();
      });
    }
  }

  void _rangeDownload(File f, int start) {
    FileMode fm = FileMode.write;
    if (start > 0) {
      fm = FileMode.append;
      this._secondaryheaders[HttpHeaders.rangeHeader] =
          "bytes=" + start.toString() + "-";
    }
    int i = 0;
    Future<HttpClientRequest> req;
    if (this._method == RemoteMethod.GET) {
      req = new HttpClient().getUrl(this._uri);
    } else {
      req = new HttpClient().postUrl(this._uri);
    }
    req.then((HttpClientRequest request) {
      this._headers.forEach((String property, String value) {
        request.headers.set(property, value);
      });
      this._secondaryheaders.forEach((String property, String value) {
        request.headers.set(property, value);
      });
      if (!this._compressedResponse) {
        request.headers
            .set(HttpHeaders.acceptEncodingHeader, 'accept-encoding=identity');
      }
      return request.close();
    }).then((HttpClientResponse response) {
      if (!this._loading && !_downloading.isClosed) {
        _downloading.add(null);
      }
      this._loading = true;
      this.response = response;
      response.listen((contents) {
        List<int> c = [];
        if (i > 0) {
          fm = FileMode.append;
        }
        if (i == 0 && this._trim) {
          c = contents.getRange(1, contents.length).toList();
          f.writeAsBytesSync(c, mode: fm);
          this._currentSize += c.length;
        } else {
          f.writeAsBytesSync(contents, mode: fm);
          this._currentSize += contents.length;
        }
        this.sp.pushState(this._currentSize, FileStateProvider.WAITING);
        i++;
      }, onDone: () {
        if (!_paused && !_download.isClosed) {
          this._secondaryheaders.remove(HttpHeaders.contentRangeHeader);
          this._secondaryheaders.remove(HttpHeaders.acceptRangesHeader);
          _downloading.close();
          this._loading = false;
          this._completed = true;
          this._totalSize = this._currentSize;
          this.sp.pushState(this._currentSize, FileStateProvider.COMPLETED);
          sp.close();
          _download.add(null);
          _download.close();
        }
        //sink.close();
        i = null;
      }, onError: (error) {
        _download.addError(error);
        _download.close();
      });
    });
  }

  //openRead implementation of a remote source
  Stream<List<int>> openRead([int start = 0, int end = 0]) {
    if (end == 0 && this._completed) {
      return this._file.openRead(start);
    } else if (end > 0 && end <= this._currentSize) {
      return this._file.openRead(start, end);
    } else if (end > 0 && this._completed) {
      return this._file.openRead(start, end);
    } else {
      return new FutureRead(this._file, this.sp, start, this.getCurrent(), end)
          .stream;
    }
  }
}
